#include "DHTesp.h"
#include <DYIRDaikin.h>

//
// for security purposes
String key = "putyoursecuritystringhere";

//
// Daikin lib:
// https://github.com/danny-source/Arduino_DY_IRDaikin

// 
// npn BC547 c transistor sheet:
// https://www.sparkfun.com/datasheets/Components/BC546.pdf

//
// DHT11, some infos
// https://chewett.co.uk/blog/1405/using-the-dht11-temperature-sensor-with-a-wemos-d1-mini-esp8266/

//
// Handles allow access control origin
// informations about handle calls from javascript from the smartphone app
// https://hackaday.io/project/12227-smart-home-in-the-iot-way/log/43430-esp8266-direct-access-from-web-app

DYIRDaikin irdaikin;


#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
// #include <ESP8266HTTPClient.h>
#include <ArduinoOTA.h>

const char* ssid     = "YOURSSID";
const char* password = "YOURPASS";


int led = LED_BUILTIN;  // wemose builtin led
int sensorPin = D4;     // sensor dht11
int daikinPin = D0;     // infra-red led

                        // AC initial assumed values
int temp = 25;          // starting assumed temperature
int statusAC = 0;       // starting assumed status: 0 off, 1 on

String pingtime="";     // var to store the ping return value for future implementations
unsigned long timer =0; // just a timer to count 3 minutes and ping


DHTesp dht;

//decode_results results;

ESP8266WebServer server(8080);

//
// blink a led to show when it's sending
// a command to the air conditioner
void ledblink(){
  pinMode(led,OUTPUT);
  digitalWrite(led, LOW);
  delay(800);
  digitalWrite(led, HIGH); // turn off build in led
}


//
// get time function from a server
// time isn't used but it could be
String getTime(String host) {
  WiFiClient client;
  while (!!!client.connect(host.c_str(), 80)) {
    Serial.println("connection failed, retrying...");
  }
  client.print("HEAD / HTTP/1.1\r\n\r\n");
  while(!!!client.available()) {
     yield();
  }
  while(client.available()){
    if (client.read() == '\n') {    
      if (client.read() == 'D') {    
        if (client.read() == 'a') {    
          if (client.read() == 't') {    
            if (client.read() == 'e') {    
              if (client.read() == ':') {    
                client.read();
                String theDate = client.readStringUntil('\r');
                client.stop();
                return theDate;
              }
            }
          }
        }
      }
    }
  }
}



//
// talk to AI to send on/off commmand with temperature setting using infra-red led
void daikinOnOff(){
  //Serial.println("statusAC: " + (String)statusAC + " temp: "+ (String)temp);
  if (statusAC == 1)  irdaikin.on(); else irdaikin.off();
  irdaikin.setSwing_off();
  irdaikin.setMode(1);
  irdaikin.setFan(4);//FAN speed to MAX
  irdaikin.setTemp( temp );
  //----everything is ok and to execute send command-----
  irdaikin.sendCommand();
  ledblink();
}



//
// set AC status, send command to Infrared, and ok message to app
void switchAcOnOff() {
   bool check = false;
   String as = "";
   int nt = 0;
   for (int i = 0; i < server.args(); i++) {
      // leggo i parametri alla ricerca di "par"
     if (server.argName(i) == "par" ) {
        as = server.arg(i);
        nt = as.toInt();
        if(nt > 1 || nt <0) nt = 0; // step range
        statusAC = nt;
     }
     if (server.argName(i) == "key" ) {
        as = server.arg(i);
        if(as == key ) check = true;
     }
   }
   server.sendHeader("Access-Control-Allow-Origin", "*"); 
   if(check) {   
     server.send(200, "application/json", "{\"ok\":\"ok\"}");
     daikinOnOff();
   } else {
     server.send(200, "application/json", "{\"ok\":\"ko\"}");    
   }
}



//
// set temperature (+1, -1) and send new settings back to app
void temperatureSet() {
   String as = "";
   int nt = 0;
   bool check = false;
   for (int i = 0; i < server.args(); i++) {
      // leggo i parametri alla ricerca di "par"
     if (server.argName(i) == "par" ) {
        as = server.arg(i);
        nt = as.toInt();
        if(nt != 1 && nt !=-1) nt = 0; // range
        temp = temp + nt;
        // temperature range limitations
        if (temp < 18) temp = 18;
        if (temp > 29) temp = 29;
     }
     // security check allows calls only for your app calls with the right key
     if (server.argName(i) == "key" ) {
        as = server.arg(i);
        if(as == key ) check = true;
     }
   }
  if(check) {   
    String s = "{\"temp\":" + (String)temp  + "}";
    server.sendHeader("Access-Control-Allow-Origin", "*");
    server.send(200, "application/json", s);  
    daikinOnOff();
  } else {
    server.send(200, "application/json", "{\"ok\":\"ko\"}");    
  }
}



//
// read data from DHT11 and send to app
void sensorData() {
   String as = "";
   bool check = false;
   for (int i = 0; i < server.args(); i++) {
     // security check allows calls only for your app calls with the right key
     if (server.argName(i) == "key" ) {
        as = server.arg(i);
        if(as == key ) check = true;
     }
   }
  if(check) { 
    int h = (int)dht.getHumidity();
    int t = (int)dht.getTemperature();
    String s = "{\"h\":" + (String)h + ",\"t\":"+(String)t  + ",\"temp\":"+(String)temp + ",\"statusAC\":"+(String)statusAC + "}";
    server.sendHeader("Access-Control-Allow-Origin", "*");
    server.send(200, "application/json", s);
  
    ledblink();
  } else {
     server.send(200, "application/json", "{\"ok\":\"ko\"}");       
  }
}



//
// handle wrong calls (404)
void handleNotFound() {
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";
  for (uint8_t i = 0; i < server.args(); i++) {
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
}



//
// connect to wifi using ssid and password
// (blink led for feedback and print . on serial monitor)
void connectToInternet() {
  pinMode(led,OUTPUT);
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  // Wait for connection
  int c=0;
  while ((WiFi.status() != WL_CONNECTED)   && c < 30) {
    c++;
    if(c%2==0) digitalWrite(led, HIGH); else digitalWrite(led, LOW);
    delay(500);
    Serial.print(".");
  }

  if (c >= 30) {
    // failed to connect. handle timeout and restart all the process
    // https://techtutorialsx.com/2017/12/29/esp8266-arduino-software-restart/
    ESP.restart();
  }
  
  digitalWrite(led, HIGH); // turn off build in led
  Serial.println("");
  Serial.print("Connected to ");
  Serial.println(ssid);
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());

}


void setup(void) {
  dht.setup(sensorPin, DHTesp::DHT11);
  Serial.begin(115200);
  Serial.println("Starting.");
  connectToInternet();
  if (MDNS.begin("condiziowifi")) { // OTA upload name
    Serial.println("MDNS responder started");
  }

  // handle these calls
  server.on("/onoff", switchAcOnOff);
  server.on("/sensor", sensorData);
  server.on("/temp", temperatureSet);
  server.onNotFound(handleNotFound);

  server.begin();
  Serial.println("HTTP server started");

  irdaikin.begin( daikinPin);


  // over the air (OTA) update setup
  ArduinoOTA.onStart([]() {
    String type;
    if (ArduinoOTA.getCommand() == U_FLASH)
      type = "sketch";
    else // U_SPIFFS
      type = "filesystem";
    // NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end()
    Serial.println("Start updating " + type);
  });
  ArduinoOTA.onEnd([]() {
    Serial.println("\nEnd");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
    else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
    else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
    else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
    else if (error == OTA_END_ERROR) Serial.println("End Failed");
  });
  ArduinoOTA.begin();



}





void loop(void) {
  server.handleClient();
  ArduinoOTA.handle();   // over the air (OTA) update loop handler


  // lost connection? reconnect
  if (WiFi.status() != WL_CONNECTED) {
     connectToInternet();
  }

  if(millis()>timer) {
    // every 3 minutes ping something to keep alive nat config on server
    pingtime = getTime("www.google.it");
    Serial.println(pingtime);
    timer = millis() + 180000;
  }

}
